@@ -6,11 +6,11 @@ from typing import Any, List
 import torch
 from torch import nn
 
-from . import layers
+from . import configs, layers
 
 
-class FMGraphNode:
-    def __init__(self, value: torch.Tensor, parents=tuple(), depth=0, layer=None):
+class Placeholder:
+    def __init__(self, value: torch.Tensor, parents=tuple(), depth=0, layer=None, batch_size_known=False):
         """Node of a Functional Model.
 
         This might represent input or intermediate values of the neural network.
@@ -26,52 +26,85 @@ class FMGraphNode:
         layer
             nn.Module object that transformed parents.
         """
-        self._v = value
+        self.v = value
         self.parents = parents
-        self.children: List[FMGraphNode] = []
+        self.children: List[Placeholder] = []
         self.layer = layer
         self.depth = depth
         self._output = None
         self._parents_outputs: List[Any] = []
-
-    @property
-    def channels(self):
-        return self._v.shape[1]
+        self.batch_size_known = batch_size_known
 
     @property
     def features(self):
-        return self._v.shape[1]
+        assert len(self.v.shape) == 2, "The data is not of [C,F] form!"
+        return self.v.shape[1]
+
+    @property
+    def C(self):
+        assert len(self.v.shape) == 4, "The data is not of [C,H,W] form!"
+        return self.v.shape[1]
+
+    @property
+    def channels(self):
+        return self.C
 
     @property
     def H(self):
-        assert len(self._v.shape) == 4, "Variable is not an image!"
-        return self._v.shape[2]
+        assert len(self.v.shape) == 4, "The data is not of [C,H,W] form!"
+        return self.v.shape[2]
 
     @property
     def W(self):
-        assert len(self._v.shape) == 4, "Variable is not an image!"
-        return self._v.shape[3]
+        assert len(self.v.shape) == 4, "The data is not of [C,H,W] form!"
+        return self.v.shape[3]
+
+    @property
+    def HW(self):
+        return (self.H, self.W)
+
+    @property
+    def CHW(self):
+        return (self.C, self.H, self.W)
+
+    @property
+    def HWC(self):
+        return (self.H, self.W, self.C)
+
+    @property
+    def batch_size(self):
+        if self.batch_size_known:
+            return self.v.shape[0]
+        else:
+            return None
 
     @property
     def shape(self):
-        return (None, *self._v.shape[1:])
+        if self.batch_size_known:
+            return self.v.shape
+        else:
+            return (None, *self.v.shape[1:])
 
     @property
     def numel(self):
-        return self._v.shape[1:].numel()
+        return self.v.shape.numel()
 
     def apply_layer(self, layer, *others):
-        if others:
-            new_depth = min(self.depth, *(o.depth for o in others)) + 1
-        else:
-            new_depth = self.depth + 1
+        all_parents = (self,) + others
+        new_depth = min(parent.depth for parent in all_parents) + 1
+        new_output = layer.__call__(self.v, *(o.v for o in others))
+        batch_size_known = all([parent.batch_size_known for parent in all_parents])
 
-        new_output = layer.__call__(self._v, *(o._v for o in others))
-
-        new_layer_node = FMGraphNode(value=new_output, parents=(self, *others), layer=layer, depth=new_depth)
-        self.children.append(new_layer_node)
-        for other in others:
-            other.children.append(new_layer_node)
+        new_layer_node = Placeholder(
+            value=new_output,
+            parents=all_parents,
+            layer=layer,
+            depth=new_depth,
+            batch_size_known=batch_size_known,
+        )
+        for parent in all_parents:
+            parent.children.append(new_layer_node)
+            logging.info(f"Added {new_layer_node} as child of {parent}")
         return new_layer_node
 
     def _get_all_nodes_below(self, layer_list):
@@ -128,7 +161,7 @@ class FMGraphNode:
         return self.apply_layer(layers.AnyOpLayer(lambda x: abs(x)))
 
     def __add__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             assert self.shape == other.shape, "Shapes do not match for the operation!"
             return self.apply_layer(layers.AddOpLayer(), other)
         else:
@@ -138,7 +171,7 @@ class FMGraphNode:
         return self.__add__(other)
 
     def __mul__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             assert self.shape == other.shape, "Shapes do not match for the operation!"
             return self.apply_layer(layers.MulOpLayer(), other)
         else:
@@ -151,66 +184,71 @@ class FMGraphNode:
         return self.apply_layer(layers.AnyOpLayer(op=lambda x: -x))
 
     def __pow__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             assert self.shape == other.shape, "Shapes do not match for the operation!"
-            return self.apply_layer(layers.AnyOpLayer(op=lambda x, y: x**y), other)
+            return self.apply_layer(layers.AnyOpLayer(op=lambda x, y: x ** y), other)
         else:
-            return self.apply_layer(layers.AnyOpLayer(op=lambda x: x**other))
+            return self.apply_layer(layers.AnyOpLayer(op=lambda x: x ** other))
 
     def __rpow__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             return other.__pow__(self)
         else:
-            return self.apply_layer(layers.AnyOpLayer(op=lambda x: other**x))
+            return self.apply_layer(layers.AnyOpLayer(op=lambda x: other ** x))
 
     def __sub__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             assert self.shape == other.shape, "Shapes do not match for the operation!"
             return self.apply_layer(layers.SubOpLayer(), other)
         else:
             return self.apply_layer(layers.AnyOpLayer(op=lambda x: x - other))
 
     def __rsub__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             return other.__sub__(self)
         else:
             return self.apply_layer(layers.AnyOpLayer(op=lambda x: other - x))
 
     def __truediv__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             assert self.shape == other.shape, "Shapes do not match for the operation!"
             return self.apply_layer(layers.AnyOpLayer(op=lambda x, y: x / y), other)
         else:
             return self.apply_layer(layers.AnyOpLayer(op=lambda x: x / other))
 
     def __rtruediv__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             return other.__truediv__(self)
         else:
             return self.apply_layer(layers.AnyOpLayer(op=lambda x: other / x))
 
     def __matmul__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             return self.apply_layer(layers.MatmulOpLayer(), other)
         else:
             return self.apply_layer(layers.AnyOpLayer(op=lambda x: x @ other))
 
     def __rmatmul__(self, other):
-        if isinstance(other, FMGraphNode):
+        if isinstance(other, Placeholder):
             return other.__matmul__(self)
         else:
             return self.apply_layer(layers.AnyOpLayer(op=lambda x: other @ x))
 
+    def __repr__(self):
+        addr = f"Placeholder at {hex(id(self))};"
+        info = f"child of {len(self.parents)}; parent of {len(self.children)}"
+        return "<" + addr + " " + info + ">"
 
-class Input(FMGraphNode):
+
+class Input(Placeholder):
     def __init__(
         self,
-        shape,
+        shape=None,
+        batch_shape=None,
         dtype=torch.float32,
         min_value=0.0,
         max_value=1.0,
-        _batch_size=1,
-        _use_tensor=None,
+        custom_tensor=None,
     ):
         """Input to the Functional Model.
 
@@ -223,7 +261,11 @@ class Input(FMGraphNode):
         Parameters
         ----------
         shape
-            Shape of the real data. Should NOT include the batch dimension.
+            Shape of the real data NOT including the batch dimension.
+        batch_shape
+            Shape of the real data including the batch dimension.
+            Should be provided instead ``shape`` if cuda graphs will be used.
+            If both ``shape`` and ``batch_shape`` are given, ``batch_shape`` has higher priority.
         dtype
             Dtype of the real data that will be the input of the network.
         min_value
@@ -233,17 +275,30 @@ class Input(FMGraphNode):
         max_value
             As above, but the maximal value.
         """
-        if _use_tensor is not None:
-            super().__init__(value=_use_tensor)
+        if custom_tensor is not None:
+            self.was_batch_size_provided = True
+            super().__init__(value=custom_tensor, batch_size_known=True)
             return
 
-        value = torch.rand(_batch_size, *shape) * (max_value - min_value) + min_value
+        if batch_shape is not None:
+            batch_size = batch_shape[0]
+            shape = batch_shape[1:]
+            batch_size_known = True
+        elif shape is not None:
+            # We use batch_size of 1 under the hood
+            # but we don't tell it to the user
+            batch_size = 1
+            batch_size_known = False
+        else:
+            raise ValueError("Shape argument is required!")
+
+        value = torch.rand(batch_size, *shape) * (max_value - min_value) + min_value
         value = value.to(dtype)
-        super().__init__(value=value)
+        super().__init__(value=value, batch_size_known=batch_size_known)
 
 
 class FunctionalModel(nn.Module):
-    def __init__(self, inputs, outputs):
+    def __init__(self, inputs, outputs, enable_cuda_graphs=False):
         """A PyTorch model that applies operations defined by the graph.
 
         All operations that changed ``inputs`` into ``outputs`` will be applied
@@ -259,14 +314,14 @@ class FunctionalModel(nn.Module):
         super().__init__()
         logging.info("Creating a Functional Model...")
 
-        if isinstance(inputs, FMGraphNode):
+        if isinstance(inputs, Placeholder):
             inputs = (inputs,)
-        assert all(isinstance(x, FMGraphNode) for x in inputs)
+        assert all(isinstance(x, Placeholder) for x in inputs)
         self.inputs = inputs
 
-        if isinstance(outputs, FMGraphNode):
+        if isinstance(outputs, Placeholder):
             outputs = (outputs,)
-        assert all(isinstance(x, FMGraphNode) for x in outputs)
+        assert all(isinstance(x, Placeholder) for x in outputs)
         self.outputs = outputs
 
         self._has_single_input = len(self.inputs) == 1
@@ -275,6 +330,14 @@ class FunctionalModel(nn.Module):
         self._prune_unused_layers()
         self._register_reachable_modules()
 
+        self.cuda_graphs_enabled = False
+
+        if enable_cuda_graphs:
+            self._enable_cuda_graphs(inputs)
+
+        if configs.MODULE_CALL_OPTIMIZATION:
+            configs.remove_call_wrapper_from_all_modules()
+
     def forward(self, inputs):
         if self._has_single_input:
             self.inputs[0]._forward_edge(inputs)
@@ -302,8 +365,27 @@ class FunctionalModel(nn.Module):
         else:
             return tuple(node.shape for node in self.outputs)
 
+    def _enable_cuda_graphs(self, inputs):
+        msg = (
+            "CUDA Graphs can result in undefined behaviour! "
+            "Please read https://pytorch.org/docs/stable/notes/cuda.html#constraints."
+        )
+        logging.warning(msg)
+        assert torch.cuda.is_available(), "CUDA acceleration is not available!"
+        for x in inputs:
+            assert x.batch_size_known, "Must provide batch size for each input!"
+
+        self.cuda()
+        input_tensors = tuple(x.v.cuda() for x in inputs)
+        torch.cuda.make_graphed_callables(self, sample_args=input_tensors)
+        self.cuda_graphs_enabled = True
+
     def _register_module(self, node):
-        if not isinstance(node.layer, nn.Module) or node.layer in self._registered_modules:
+        if not isinstance(node.layer, nn.Module):
+            logging.info(f"Not registering {node.layer} (not a nn.Module)!")
+            return False
+        elif node.layer in self._registered_modules:
+            logging.info(f"Not registering {node.layer} (already registered)!")
             return False
 
         num_modules = len(self._registered_modules)
